@@ -56,6 +56,18 @@ parse_response_headers (SoupMessage *req,
 	if (*encoding == SOUP_ENCODING_UNRECOGNIZED)
 		return SOUP_STATUS_MALFORMED;
 
+        /* mirror Mozilla here:
+         * (see netwerk/protocol/http/src/nsHttpTransaction.cpp for details)
+         *
+         * HTTP servers have been known to send erroneous Content-Length headers.
+         * So, unless the connection is persistent, we must make allowances for a
+         * possibly invalid Content-Length header. Thus, if NOT persistent, we 
+         * simply accept the whole message.
+         */
+        if (*encoding == SOUP_ENCODING_CONTENT_LENGTH &&
+            !soup_message_is_keepalive (req))
+                *encoding = SOUP_ENCODING_EOF;
+
 	return SOUP_STATUS_OK;
 }
 
@@ -92,7 +104,7 @@ get_request_headers (SoupMessage *req, GString *header,
 	} else {
 		g_string_append_printf (header, "%s %s HTTP/1.1\r\n",
 					req->method, uri_string);
-		if (!soup_message_headers_get (req->request_headers, "Host")) {
+		if (!soup_message_headers_get_one (req->request_headers, "Host")) {
 			if (soup_uri_uses_default_port (uri)) {
 				g_string_append_printf (header, "Host: %s\r\n",
 							uri_host);
@@ -109,7 +121,8 @@ get_request_headers (SoupMessage *req, GString *header,
 	*encoding = soup_message_headers_get_encoding (req->request_headers);
 	if ((*encoding == SOUP_ENCODING_CONTENT_LENGTH ||
 	     *encoding == SOUP_ENCODING_NONE) &&
-	    req->request_body->length > 0 &&
+	    (req->request_body->length > 0 ||
+	     soup_message_headers_get_one (req->request_headers, "Content-Type")) &&
 	    !soup_message_headers_get_content_length (req->request_headers)) {
 		*encoding = SOUP_ENCODING_CONTENT_LENGTH;
 		soup_message_headers_set_content_length (req->request_headers,
