@@ -8,6 +8,7 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include "soup-address.h"
 #include "soup-auth.h"
 #include "soup-enum-types.h"
 #include "soup-marshal.h"
@@ -51,6 +52,17 @@
  *
  * Represents an HTTP message being sent or received.
  *
+ * @status_code will normally be a #SoupKnownStatusCode, eg,
+ * %SOUP_STATUS_OK, though of course it might actually be an unknown
+ * status code. @reason_phrase is the actual text returned from the
+ * server, which may or may not correspond to the "standard"
+ * description of @status_code. At any rate, it is almost certainly
+ * not localized, and not very descriptive even if it is in the user's
+ * language; you should not use @reason_phrase in user-visible
+ * messages. Rather, you should look at @status_code, and determine an
+ * end-user-appropriate message based on that and on what you were
+ * trying to do.
+ *
  * As described in the #SoupMessageBody documentation, the
  * @request_body and @response_body %data fields will not necessarily
  * be filled in at all times. When they are filled in, they will be
@@ -87,6 +99,7 @@ enum {
 	GOT_HEADERS,
 	GOT_CHUNK,
 	GOT_BODY,
+	CONTENT_SNIFFED,
 
 	RESTARTED,
 	FINISHED,
@@ -145,12 +158,16 @@ finalize (GObject *object)
 
 	if (priv->uri)
 		soup_uri_free (priv->uri);
+	if (priv->addr)
+		g_object_unref (priv->addr);
 
 	if (priv->auth)
 		g_object_unref (priv->auth);
 	if (priv->proxy_auth)
 		g_object_unref (priv->proxy_auth);
 
+	g_slist_free (priv->disabled_features);
+
 	soup_message_body_free (msg->request_body);
 	soup_message_headers_free (msg->request_headers);
 	soup_message_body_free (msg->response_body);
@@ -247,6 +264,8 @@ soup_message_class_init (SoupMessageClass *message_class)
 	 * Unlike #SoupMessage::wrote_chunk, this is emitted after
 	 * every successful write() call, not only after finishing a
 	 * complete "chunk".
+	 *
+	 * Since: 2.4.1
 	 **/
 	signals[WROTE_BODY_DATA] =
 		g_signal_new ("wrote_body_data",
@@ -385,6 +404,44 @@ soup_message_class_init (SoupMessageClass *message_class)
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
 
+	/**
+	 * SoupMessage::content-sniffed:
+	 * @msg: the message
+	 * @type: the content type that we got from sniffing
+	 * @params: a #GHashTable with the parameters
+	 *
+	 * This signal is emitted after %got-headers, and before the
+	 * first %got-chunk. If content sniffing is disabled, or no
+	 * content sniffing will be performed, due to the sniffer
+	 * deciding to trust the Content-Type sent by the server, this
+	 * signal is emitted immediately after %got_headers, and @type
+	 * is %NULL.
+	 *
+	 * If the #SoupContentSniffer feature is enabled, and the
+	 * sniffer decided to perform sniffing, the first %got_chunk
+	 * emission may be delayed, so that the sniffer has enough
+	 * data to correctly sniff the content. It notified the
+	 * library user that the content has been sniffed, and allows
+	 * it to change the header contents in the message, if
+	 * desired.
+	 *
+	 * After this signal is emitted, the data that was spooled so
+	 * that sniffing could be done is delivered on the first
+	 * emission of %got_chunk.
+	 *
+	 * Since: 2.27.3
+	 **/
+	signals[CONTENT_SNIFFED] =
+		g_signal_new ("content_sniffed",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_FIRST,
+			      0,
+			      NULL, NULL,
+			      soup_marshal_NONE__STRING_BOXED,
+			      G_TYPE_NONE, 2,
+			      G_TYPE_STRING,
+			      G_TYPE_HASH_TABLE);
+
 	/**
 	 * SoupMessage::restarted:
 	 * @msg: the message
@@ -421,6 +478,12 @@ soup_message_class_init (SoupMessageClass *message_class)
 			      G_TYPE_NONE, 0);
 
 	/* properties */
+	/**
+	 * SOUP_MESSAGE_METHOD:
+	 *
+	 * Alias for the #SoupMessage:method property. (The message's
+	 * HTTP method.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_METHOD,
 		g_param_spec_string (SOUP_MESSAGE_METHOD,
@@ -428,6 +491,12 @@ soup_message_class_init (SoupMessageClass *message_class)
 				     "The message's HTTP method",
 				     SOUP_METHOD_GET,
 				     G_PARAM_READWRITE));
+	/**
+	 * SOUP_MESSAGE_URI:
+	 *
+	 * Alias for the #SoupMessage:uri property. (The message's
+	 * #SoupURI.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_URI,
 		g_param_spec_boxed (SOUP_MESSAGE_URI,
@@ -435,6 +504,12 @@ soup_message_class_init (SoupMessageClass *message_class)
 				    "The message's Request-URI",
 				    SOUP_TYPE_URI,
 				    G_PARAM_READWRITE));
+	/**
+	 * SOUP_MESSAGE_HTTP_VERSION:
+	 *
+	 * Alias for the #SoupMessage:http-version property. (The
+	 * message's #SoupHTTPVersion.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_HTTP_VERSION,
 		g_param_spec_enum (SOUP_MESSAGE_HTTP_VERSION,
@@ -443,6 +518,12 @@ soup_message_class_init (SoupMessageClass *message_class)
 				   SOUP_TYPE_HTTP_VERSION,
 				   SOUP_HTTP_1_1,
 				   G_PARAM_READWRITE));
+	/**
+	 * SOUP_MESSAGE_FLAGS:
+	 *
+	 * Alias for the #SoupMessage:flags property. (The message's
+	 * #SoupMessageFlags.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_FLAGS,
 		g_param_spec_flags (SOUP_MESSAGE_FLAGS,
@@ -451,6 +532,12 @@ soup_message_class_init (SoupMessageClass *message_class)
 				    SOUP_TYPE_MESSAGE_FLAGS,
 				    0,
 				    G_PARAM_READWRITE));
+	/**
+	 * SOUP_MESSAGE_SERVER_SIDE:
+	 *
+	 * Alias for the #SoupMessage:server-side property. (%TRUE if
+	 * the message was created by #SoupServer.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_SERVER_SIDE,
 		g_param_spec_boolean (SOUP_MESSAGE_SERVER_SIDE,
@@ -458,6 +545,12 @@ soup_message_class_init (SoupMessageClass *message_class)
 				      "Whether or not the message is server-side rather than client-side",
 				      FALSE,
 				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_MESSAGE_STATUS_CODE:
+	 *
+	 * Alias for the #SoupMessage:status-code property. (The
+	 * message's HTTP response status code.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_STATUS_CODE,
 		g_param_spec_uint (SOUP_MESSAGE_STATUS_CODE,
@@ -465,6 +558,12 @@ soup_message_class_init (SoupMessageClass *message_class)
 				   "The HTTP response status code",
 				   0, 599, 0,
 				   G_PARAM_READWRITE));
+	/**
+	 * SOUP_MESSAGE_REASON_PHRASE:
+	 *
+	 * Alias for the #SoupMessage:reason-phrase property. (The
+	 * message's HTTP response reason phrase.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_REASON_PHRASE,
 		g_param_spec_string (SOUP_MESSAGE_REASON_PHRASE,
@@ -496,6 +595,10 @@ set_property (GObject *object, guint prop_id,
 		break;
 	case PROP_SERVER_SIDE:
 		priv->server_side = g_value_get_boolean (value);
+		if (priv->server_side) {
+			soup_message_headers_set_encoding (msg->response_headers,
+							   SOUP_ENCODING_CONTENT_LENGTH);
+		}
 		break;
 	case PROP_STATUS_CODE:
 		soup_message_set_status (msg, g_value_get_uint (value));
@@ -796,6 +899,24 @@ soup_message_got_body (SoupMessage *msg)
 	g_signal_emit (msg, signals[GOT_BODY], 0);
 }
 
+/**
+ * soup_message_content_sniffed:
+ * @msg: a #SoupMessage
+ * @type: a string with the sniffed content type
+ * @params: a #GHashTable with the parameters
+ *
+ * Emits the %content_sniffed signal, indicating that the IO layer
+ * finished sniffing the content type for @msg. If content sniffing
+ * will not be performed, due to the sniffer deciding to trust the
+ * Content-Type sent by the server, this signal is emitted immediately
+ * after %got_headers, with %NULL as @content_type.
+ **/
+void
+soup_message_content_sniffed (SoupMessage *msg, const char *content_type, GHashTable *params)
+{
+	g_signal_emit (msg, signals[CONTENT_SNIFFED], 0, content_type, params);
+}
+
 static void
 restarted (SoupMessage *req)
 {
@@ -834,7 +955,10 @@ finished (SoupMessage *req)
 void
 soup_message_finished (SoupMessage *msg)
 {
-	g_signal_emit (msg, signals[FINISHED], 0);
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+
+	if (priv->io_status != SOUP_MESSAGE_IO_STATUS_FINISHED)
+		g_signal_emit (msg, signals[FINISHED], 0);
 }
 
 static void
@@ -857,7 +981,7 @@ header_handler_metamarshal (GClosure *closure, GValue *return_value,
 		return;
 
 	hdrs = priv->server_side ? msg->request_headers : msg->response_headers;
-	if (soup_message_headers_get (hdrs, header_name)) {
+	if (soup_message_headers_get_one (hdrs, header_name)) {
 		closure->marshal (closure, return_value, n_param_values,
 				  param_values, invocation_hint,
 				  ((GCClosure *)closure)->callback);
@@ -1098,6 +1222,10 @@ soup_message_cleanup_response (SoupMessage *req)
 
 	soup_message_body_truncate (req->response_body);
 	soup_message_headers_clear (req->response_headers);
+	if (priv->server_side) {
+		soup_message_headers_set_encoding (req->response_headers,
+						   SOUP_ENCODING_CONTENT_LENGTH);
+	}
 
 	req->status_code = SOUP_STATUS_NONE;
 	if (req->reason_phrase) {
@@ -1227,13 +1355,19 @@ soup_message_is_keepalive (SoupMessage *msg)
 {
 	const char *c_conn, *s_conn;
 
-	c_conn = soup_message_headers_get (msg->request_headers, "Connection");
-	s_conn = soup_message_headers_get (msg->response_headers, "Connection");
+	c_conn = soup_message_headers_get_list (msg->request_headers,
+						"Connection");
+	s_conn = soup_message_headers_get_list (msg->response_headers,
+						"Connection");
 
 	if (msg->status_code == SOUP_STATUS_OK &&
 	    msg->method == SOUP_METHOD_CONNECT)
 		return TRUE;
 
+	/* Not persistent if the server sent a terminate-by-EOF response */
+	if (soup_message_headers_get_encoding (msg->response_headers) == SOUP_ENCODING_EOF)
+		return FALSE;
+
 	if (SOUP_MESSAGE_GET_PRIVATE (msg)->http_version == SOUP_HTTP_1_0) {
 		/* Only persistent if the client requested keepalive
 		 * and the server agreed.
@@ -1241,11 +1375,9 @@ soup_message_is_keepalive (SoupMessage *msg)
 
 		if (!c_conn || !s_conn)
 			return FALSE;
-		if (soup_header_contains (c_conn, "Keep-Alive") ||
-		    soup_header_contains (s_conn, "Keep-Alive"))
+		if (!soup_header_contains (c_conn, "Keep-Alive") ||
+		    !soup_header_contains (s_conn, "Keep-Alive"))
 			return FALSE;
-
-		return TRUE;
 	} else {
 		/* Normally persistent unless either side requested otherwise */
 		if (c_conn && soup_header_contains (c_conn, "close"))
@@ -1253,12 +1385,10 @@ soup_message_is_keepalive (SoupMessage *msg)
 		if (s_conn && soup_header_contains (s_conn, "close"))
 			return FALSE;
 
-		/* But not if the server sent a terminate-by-EOF response */
-		if (soup_message_headers_get_encoding (msg->response_headers) == SOUP_ENCODING_EOF)
-			return FALSE;
-
 		return TRUE;
 	}
+
+	return TRUE;
 }
 
 /**
@@ -1280,6 +1410,10 @@ soup_message_set_uri (SoupMessage *msg, SoupURI *uri)
 
 	if (priv->uri)
 		soup_uri_free (priv->uri);
+	if (priv->addr) {
+		g_object_unref (priv->addr);
+		priv->addr = NULL;
+	}
 	priv->uri = soup_uri_copy (uri);
 
 	g_object_notify (G_OBJECT (msg), SOUP_MESSAGE_URI);
@@ -1301,6 +1435,33 @@ soup_message_get_uri (SoupMessage *msg)
 	return SOUP_MESSAGE_GET_PRIVATE (msg)->uri;
 }
 
+/**
+ * soup_message_get_address:
+ * @msg: a #SoupMessage
+ *
+ * Gets the address @msg's URI points to. After first setting the
+ * URI on a message, this will be unresolved, although the message's
+ * session will resolve it before sending the message.
+ *
+ * Return value: the address @msg's URI points to
+ *
+ * Since: 2.26
+ **/
+SoupAddress *
+soup_message_get_address (SoupMessage *msg)
+{
+	SoupMessagePrivate *priv;
+
+	g_return_val_if_fail (SOUP_IS_MESSAGE (msg), NULL);
+
+	priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+	if (!priv->addr) {
+		priv->addr = soup_address_new (priv->uri->host,
+					       priv->uri->port);
+	}
+	return priv->addr;
+}
+
 /**
  * soup_message_set_status:
  * @msg: a #SoupMessage
@@ -1442,3 +1603,53 @@ soup_message_set_chunk_allocator (SoupMessage *msg,
 	priv->chunk_allocator_data    = user_data;
 	priv->chunk_allocator_dnotify = destroy_notify;
 }
+
+/**
+ * soup_message_disable_feature:
+ * @msg: a #SoupMessage
+ * @feature_type: the #GType of a #SoupSessionFeature
+ *
+ * This disables the actions of #SoupSessionFeature<!-- -->s with the
+ * given @feature_type (or a subclass of that type) on @msg, so that
+ * @msg is processed as though the feature(s) hadn't been added to the
+ * session. Eg, passing #SOUP_TYPE_PROXY_RESOLVER for @feature_type
+ * will disable proxy handling and cause @msg to be sent directly to
+ * the indicated origin server, regardless of system proxy
+ * configuration.
+ *
+ * You must call this before queueing @msg on a session; calling it on
+ * a message that has already been queued is undefined. In particular,
+ * you cannot call this on a message that is being requeued after a
+ * redirect or authentication.
+ *
+ * Since: 2.28
+ **/
+void
+soup_message_disable_feature (SoupMessage *msg, GType feature_type)
+{
+	SoupMessagePrivate *priv;
+
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+
+	priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+
+	priv->disabled_features = g_slist_prepend (priv->disabled_features,
+						   GSIZE_TO_POINTER (feature_type));
+}
+
+gboolean
+soup_message_disables_feature (SoupMessage *msg, gpointer feature)
+{
+	SoupMessagePrivate *priv;
+	GSList *f;
+
+	g_return_val_if_fail (SOUP_IS_MESSAGE (msg), FALSE);
+
+	priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+
+	for (f = priv->disabled_features; f; f = f->next) {
+		if (G_TYPE_CHECK_INSTANCE_TYPE (feature, (GType) GPOINTER_TO_SIZE (f->data)))
+			return TRUE;
+	}
+	return FALSE;
+}
