@@ -22,7 +22,7 @@ typedef void (*SoupSessionCallback) (SoupSession           *session,
 				     SoupMessage           *msg,
 				     gpointer               user_data);
 
-struct SoupSession {
+struct _SoupSession {
 	GObject parent;
 
 };
@@ -61,8 +61,12 @@ GType soup_session_get_type (void);
 #define SOUP_SESSION_USE_NTLM           "use-ntlm"
 #define SOUP_SESSION_SSL_CA_FILE        "ssl-ca-file"
 #define SOUP_SESSION_ASYNC_CONTEXT      "async-context"
-#define SOUP_SESSION_TIMEOUT		"timeout"
+#define SOUP_SESSION_TIMEOUT            "timeout"
 #define SOUP_SESSION_USER_AGENT         "user-agent"
+#define SOUP_SESSION_IDLE_TIMEOUT       "idle-timeout"
+#define SOUP_SESSION_ADD_FEATURE            "add-feature"
+#define SOUP_SESSION_ADD_FEATURE_BY_TYPE    "add-feature-by-type"
+#define SOUP_SESSION_REMOVE_FEATURE_BY_TYPE "remove-feature-by-type"
 
 GMainContext   *soup_session_get_async_context(SoupSession           *session);
 
@@ -86,6 +90,23 @@ void            soup_session_cancel_message   (SoupSession           *session,
 					       guint                  status_code);
 void            soup_session_abort            (SoupSession           *session);
 
+
+void                soup_session_add_feature            (SoupSession        *session,
+							 SoupSessionFeature *feature);
+void                soup_session_add_feature_by_type    (SoupSession        *session,
+							 GType               feature_type);
+void                soup_session_remove_feature         (SoupSession        *session,
+							 SoupSessionFeature *feature);
+void                soup_session_remove_feature_by_type (SoupSession        *session,
+							 GType               feature_type);
+GSList             *soup_session_get_features           (SoupSession        *session,
+							 GType               feature_type);
+SoupSessionFeature *soup_session_get_feature            (SoupSession        *session,
+							 GType               feature_type);
+SoupSessionFeature *soup_session_get_feature_for_message(SoupSession        *session,
+							 GType               feature_type,
+							 SoupMessage        *msg);
+
 G_END_DECLS
 
 #endif /* SOUP_SESSION_H */
