@@ -38,7 +38,9 @@
  *
  * A #SoupURI represents a (parsed) URI. #SoupURI supports RFC 3986
  * (URI Generic Syntax), and can parse any valid URI. However, libsoup
- * only uses "http" and "https" URIs internally.
+ * only uses "http" and "https" URIs internally; You can use
+ * SOUP_URI_VALID_FOR_HTTP() to test if a #SoupURI is a valid HTTP
+ * URI.
  *
  * @scheme will always be set in any URI. It is an interned string and
  * is always all lowercase. (If you parse a URI with a non-lowercase
@@ -77,25 +79,38 @@
  * as well, but this would be more annoying than useful.)
  **/
 
+/**
+ * SOUP_URI_VALID_FOR_HTTP:
+ * @uri: a #SoupURI
+ *
+ * Tests if @uri is a valid #SoupURI for HTTP communication; that is, if
+ * it can be used to construct a #SoupMessage.
+ *
+ * Return value: %TRUE if @uri is a valid "http" or "https" URI.
+ *
+ * Since: 2.24
+ **/
+
 static void append_uri_encoded (GString *str, const char *in, const char *extra_enc_chars);
 static char *uri_decoded_copy (const char *str, int length);
-static char *uri_normalized_copy (const char *str, int length, const char *unescape_extra);
+static char *uri_normalized_copy (const char *str, int length, const char *unescape_extra, gboolean fixup);
 
-const char *_SOUP_URI_SCHEME_HTTP, *_SOUP_URI_SCHEME_HTTPS;
+gpointer _SOUP_URI_SCHEME_HTTP, _SOUP_URI_SCHEME_HTTPS;
 
 static inline const char *
 soup_uri_get_scheme (const char *scheme, int len)
 {
-	if (len == 4 && !strncmp (scheme, "http", 4)) {
+	if (len == 4 && !g_ascii_strncasecmp (scheme, "http", len)) {
 		return SOUP_URI_SCHEME_HTTP;
-	} else if (len == 5 && !strncmp (scheme, "https", 5)) {
+	} else if (len == 5 && !g_ascii_strncasecmp (scheme, "https", len)) {
 		return SOUP_URI_SCHEME_HTTPS;
 	} else {
 		char *lower_scheme;
 
 		lower_scheme = g_ascii_strdown (scheme, len);
-		scheme = g_intern_string (lower_scheme);
-		g_free (lower_scheme);
+		scheme = g_intern_static_string (lower_scheme);
+		if (scheme != (const char *)lower_scheme)
+			g_free (lower_scheme);
 		return scheme;
 	}
 }
@@ -138,7 +153,7 @@ soup_uri_new_with_base (SoupURI *base, const char *uri_string)
 	end = hash = strchr (uri_string, '#');
 	if (hash && hash[1]) {
 		uri->fragment = uri_normalized_copy (hash + 1, strlen (hash + 1),
-						     NULL);
+						     NULL, FALSE);
 		if (!uri->fragment) {
 			soup_uri_free (uri);
 			return NULL;
@@ -235,7 +250,7 @@ soup_uri_new_with_base (SoupURI *base, const char *uri_string)
 		if (question[1]) {
 			uri->query = uri_normalized_copy (question + 1,
 							  end - (question + 1),
-							  NULL);
+							  NULL, TRUE);
 			if (!uri->query) {
 				soup_uri_free (uri);
 				return NULL;
@@ -246,7 +261,7 @@ soup_uri_new_with_base (SoupURI *base, const char *uri_string)
 
 	if (end != uri_string) {
 		uri->path = uri_normalized_copy (uri_string, end - uri_string,
-						 NULL);
+						 NULL, TRUE);
 		if (!uri->path) {
 			soup_uri_free (uri);
 			return NULL;
@@ -337,7 +352,7 @@ soup_uri_new_with_base (SoupURI *base, const char *uri_string)
 	/* HTTP-specific stuff */
 	if (uri->scheme == SOUP_URI_SCHEME_HTTP ||
 	    uri->scheme == SOUP_URI_SCHEME_HTTPS) {
-		if (!uri->host) {
+		if (!SOUP_URI_VALID_FOR_HTTP (uri)) {
 			soup_uri_free (uri);
 			return NULL;
 		}
@@ -360,7 +375,9 @@ soup_uri_new_with_base (SoupURI *base, const char *uri_string)
  * Parses an absolute URI.
  *
  * You can also pass %NULL for @uri_string if you want to get back an
- * "empty" #SoupURI that you can fill in by hand.
+ * "empty" #SoupURI that you can fill in by hand. (You will need to
+ * call at least soup_uri_set_scheme() and soup_uri_set_path(), since
+ * those fields are required.)
  *
  * Return value: a #SoupURI, or %NULL.
  **/
@@ -403,6 +420,8 @@ soup_uri_to_string (SoupURI *uri, gboolean just_path_and_query)
 	GString *str;
 	char *return_result;
 
+	g_return_val_if_fail (uri != NULL, NULL);
+
 	/* IF YOU CHANGE ANYTHING IN THIS FUNCTION, RUN
 	 * tests/uri-parsing AFTERWARD.
 	 */
@@ -410,7 +429,7 @@ soup_uri_to_string (SoupURI *uri, gboolean just_path_and_query)
 	str = g_string_sized_new (20);
 
 	if (uri->scheme && !just_path_and_query)
-		g_string_sprintfa (str, "%s:", uri->scheme);
+		g_string_append_printf (str, "%s:", uri->scheme);
 	if (uri->host && !just_path_and_query) {
 		g_string_append (str, "//");
 		if (uri->user) {
@@ -475,49 +494,6 @@ soup_uri_copy (SoupURI *uri)
 	return dup;
 }
 
-/* Temporarily still used by SoupSession, but no longer public */
-SoupURI *soup_uri_copy_root (SoupURI *uri);
-gboolean soup_uri_host_equal (gconstpointer v1, gconstpointer v2);
-guint    soup_uri_host_hash (gconstpointer key);
-
-SoupURI *
-soup_uri_copy_root (SoupURI *uri)
-{
-	SoupURI *dup;
-
-	g_return_val_if_fail (uri != NULL, NULL);
-
-	dup = g_slice_new0 (SoupURI);
-	dup->scheme = uri->scheme;
-	dup->host   = g_strdup (uri->host);
-	dup->port   = uri->port;
-
-	return dup;
-}
-
-guint
-soup_uri_host_hash (gconstpointer key)
-{
-	const SoupURI *uri = key;
-
-	return GPOINTER_TO_UINT (uri->scheme) + uri->port +
-		soup_str_case_hash (uri->host);
-}
-
-gboolean
-soup_uri_host_equal (gconstpointer v1, gconstpointer v2)
-{
-	const SoupURI *one = v1;
-	const SoupURI *two = v2;
-
-	if (one->scheme != two->scheme)
-		return FALSE;
-	if (one->port != two->port)
-		return FALSE;
-
-	return g_ascii_strcasecmp (one->host, two->host) == 0;
-}
-
 static inline gboolean
 parts_equal (const char *one, const char *two, gboolean insensitive)
 {
@@ -679,10 +655,12 @@ soup_uri_decode (const char *part)
 }
 
 static char *
-uri_normalized_copy (const char *part, int length, const char *unescape_extra)
+uri_normalized_copy (const char *part, int length,
+		     const char *unescape_extra, gboolean fixup)
 {
 	unsigned char *s, *d, c;
 	char *normalized = g_strndup (part, length);
+	gboolean need_fixup = FALSE;
 
 	s = d = (unsigned char *)normalized;
 	do {
@@ -703,10 +681,27 @@ uri_normalized_copy (const char *part, int length, const char *unescape_extra)
 				*d++ = g_ascii_toupper (*s++);
 				*d++ = g_ascii_toupper (*s);
 			}
-		} else
+		} else {
+			if (*s == ' ')
+				need_fixup = TRUE;
 			*d++ = *s;
+		}
 	} while (*s++);
 
+	if (fixup && need_fixup) {
+		char *tmp, *sp;
+		/* This code is lame, but so are people who put
+		 * unencoded spaces in URLs!
+		 */
+		while ((sp = strchr (normalized, ' '))) {
+			tmp = g_strdup_printf ("%.*s%%20%s",
+					       (int)(sp - normalized),
+					       normalized, sp + 1);
+			g_free (normalized);
+			normalized = tmp;
+		};
+	}
+
 	return normalized;
 }
 
@@ -734,7 +729,7 @@ uri_normalized_copy (const char *part, int length, const char *unescape_extra)
 char *
 soup_uri_normalize (const char *part, const char *unescape_extra)
 {
-	return uri_normalized_copy (part, strlen (part), unescape_extra);
+	return uri_normalized_copy (part, strlen (part), unescape_extra, FALSE);
 }
 
 
@@ -927,6 +922,84 @@ soup_uri_set_fragment (SoupURI *uri, const char *fragment)
 	uri->fragment = g_strdup (fragment);
 }
 
+/**
+ * soup_uri_copy_host:
+ * @uri: a #SoupUri
+ *
+ * Makes a copy of @uri, considering only the protocol, host, and port
+ *
+ * Return value: the new #SoupUri
+ *
+ * Since: 2.26.3
+ **/
+SoupURI *
+soup_uri_copy_host (SoupURI *uri)
+{
+	SoupURI *dup;
+
+	g_return_val_if_fail (uri != NULL, NULL);
+
+	dup = soup_uri_new (NULL);
+	dup->scheme = uri->scheme;
+	dup->host   = g_strdup (uri->host);
+	dup->port   = uri->port;
+	if (dup->scheme == SOUP_URI_SCHEME_HTTP ||
+	    dup->scheme == SOUP_URI_SCHEME_HTTPS)
+		dup->path = g_strdup ("");
+
+	return dup;
+}
+
+/**
+ * soup_uri_host_hash:
+ * @key: a #SoupURI
+ *
+ * Hashes @key, considering only the scheme, host, and port.
+ *
+ * Return value: a hash
+ *
+ * Since: 2.26.3
+ **/
+guint
+soup_uri_host_hash (gconstpointer key)
+{
+	const SoupURI *uri = key;
+
+	g_return_val_if_fail (uri != NULL && uri->host != NULL, 0);
+
+	return GPOINTER_TO_UINT (uri->scheme) + uri->port +
+		soup_str_case_hash (uri->host);
+}
+
+/**
+ * soup_uri_host_equal:
+ * @v1: a #SoupURI
+ * @v2: a #SoupURI
+ *
+ * Compares @v1 and @v2, considering only the scheme, host, and port.
+ *
+ * Return value: whether or not the URIs are equal in scheme, host,
+ * and port.
+ *
+ * Since: 2.26.3
+ **/
+gboolean
+soup_uri_host_equal (gconstpointer v1, gconstpointer v2)
+{
+	const SoupURI *one = v1;
+	const SoupURI *two = v2;
+
+	g_return_val_if_fail (one != NULL && two != NULL, one == two);
+	g_return_val_if_fail (one->host != NULL && two->host != NULL, one->host == two->host);
+
+	if (one->scheme != two->scheme)
+		return FALSE;
+	if (one->port != two->port)
+		return FALSE;
+
+	return g_ascii_strcasecmp (one->host, two->host) == 0;
+}
+
 
 GType
 soup_uri_get_type (void)
