@@ -16,14 +16,14 @@ typedef struct {
 	const guint final_status;
 } SoupProxyTest;
 
-SoupProxyTest tests[] = {
+static SoupProxyTest tests[] = {
 	{ "GET -> 200", "", SOUP_STATUS_OK },
 	{ "GET -> 404", "/not-found", SOUP_STATUS_NOT_FOUND },
 	{ "GET -> 401 -> 200", "/Basic/realm1/", SOUP_STATUS_OK },
 	{ "GET -> 401 -> 401", "/Basic/realm2/", SOUP_STATUS_UNAUTHORIZED },
-	{ "GET -> 403", "http://no-proxy.example.com/", SOUP_STATUS_FORBIDDEN },
+	{ "GET -> 403", "http://no-such-hostname.xx/", SOUP_STATUS_FORBIDDEN },
 };
-int ntests = sizeof (tests) / sizeof (tests[0]);
+static int ntests = sizeof (tests) / sizeof (tests[0]);
 
 #define HTTP_SERVER    "http://127.0.0.1:47524"
 #define HTTPS_SERVER   "https://127.0.0.1:47525"
@@ -48,6 +48,22 @@ static void
 authenticate (SoupSession *session, SoupMessage *msg,
 	      SoupAuth *auth, gboolean retrying, gpointer data)
 {
+	if (msg->status_code == SOUP_STATUS_UNAUTHORIZED) {
+		if (soup_auth_is_for_proxy (auth)) {
+			debug_printf (1, "  got proxy auth object for 401!\n");
+			errors++;
+		}
+	} else if (msg->status_code == SOUP_STATUS_PROXY_UNAUTHORIZED) {
+		if (!soup_auth_is_for_proxy (auth)) {
+			debug_printf (1, "  got regular auth object for 407!\n");
+			errors++;
+		}
+	} else {
+		debug_printf (1, "  got authenticate signal with status %d\n",
+			      msg->status_code);
+		errors++;
+	}
+
 	if (!retrying)
 		soup_auth_authenticate (auth, "user1", "realm1");
 }
@@ -67,9 +83,9 @@ test_url (const char *url, int proxy, guint expected, gboolean sync)
 	 * connections/auth aren't cached between tests.
 	 */
 	proxy_uri = soup_uri_new (proxies[proxy]);
-	session = g_object_new (sync ? SOUP_TYPE_SESSION_SYNC : SOUP_TYPE_SESSION_ASYNC,
-				SOUP_SESSION_PROXY_URI, proxy_uri,
-				NULL);
+	session = soup_test_session_new (sync ? SOUP_TYPE_SESSION_SYNC : SOUP_TYPE_SESSION_ASYNC,
+					 SOUP_SESSION_PROXY_URI, proxy_uri,
+					 NULL);
 	soup_uri_free (proxy_uri);
 	g_signal_connect (session, "authenticate",
 			  G_CALLBACK (authenticate), NULL);
@@ -109,15 +125,15 @@ run_test (int i, gboolean sync)
 		https_url = g_strconcat (HTTPS_SERVER, tests[i].url, NULL);
 	}
 	test_url (http_url, SIMPLE_PROXY, tests[i].final_status, sync);
-#if HAVE_SSL
+#ifdef HAVE_SSL
 	test_url (https_url, SIMPLE_PROXY, tests[i].final_status, sync);
 #endif
 	test_url (http_url, AUTH_PROXY, tests[i].final_status, sync);
-#if HAVE_SSL
+#ifdef HAVE_SSL
 	test_url (https_url, AUTH_PROXY, tests[i].final_status, sync);
 #endif
 	test_url (http_url, UNAUTH_PROXY, tests[i].final_status, sync);
-#if HAVE_SSL
+#ifdef HAVE_SSL
 	test_url (https_url, UNAUTH_PROXY, tests[i].final_status, sync);
 #endif
 
